/* velha.c */

#include <math.h>
#include <stdlib.h>
#include "velha.h"
#include "interfac.h"

/* inicie: Preenche o tabuleiro com casas vazias para retirar possiveis
   lixos.
   Recebe: Uma estrutura tabuleiro por referencia.
   Retorna: Esta estrutura preenchida com casas vazias. */

int inicie (tipo_tabuleiro *tabl)
{

  int linha,
      coluna;

  for (linha = 0; linha < 3; linha++)
    for (coluna = 0; coluna < 3; coluna++)
      (*tabl).tabuleiro [linha] [coluna] = VAZIO;
}

/* copie: Produz uma copia de um tabuleiro.
   Recebe: O tabuleiro de origem e o tabuleiro destino passado por
   referencia.
   Retorna: O tabuleiro destino com os valore iguais ao de origem. */

void copie (tipo_tabuleiro origem, tipo_tabuleiro *destino)
{

  int linha,
      coluna;

  for (linha = 0; linha < 3; linha++)
    for (coluna = 0; coluna < 3; coluna++)
      (*destino).tabuleiro [linha] [coluna] =
      origem.tabuleiro [linha] [coluna];
}

void coordenadas_movimento (int *linha, int *coluna, tipo_tabuleiro atual,
			    tipo_tabuleiro movimento)
{
  int cont_linha,
      cont_coluna;

  for (cont_linha = 0; cont_linha < 3; cont_linha++)
    for (cont_coluna = 0; cont_coluna < 3; cont_coluna++)
      if (atual.tabuleiro [cont_linha] [cont_coluna] !=
	  movimento.tabuleiro [cont_linha] [cont_coluna]) {
	*linha = ++cont_linha;
	*coluna = ++cont_coluna;
	return;
      }
  *linha = *coluna = -1;
}

int cheio (tipo_tabuleiro atual)
{

  int linha,
      coluna;

  for (linha = 0; linha < 3; linha++)
    for (coluna = 0; coluna < 3; coluna++)
      if (atual.tabuleiro [linha] [coluna] == VAZIO)
	return FALSE;
  return TRUE;

}

/* venceu: Verifica se um determinado jogador possui um posicao vencedora
   no tabuleiro.
   Recebe: O tabuleiro e a indicacao de qual jogador a ser verificado.
   Retorna: TRUE se ha uma posicao vencedora, False caso contrario. */

int venceu (int vez, tipo_tabuleiro atual)
{

  int linha,
      coluna;

  /* Verificar as linhas */

  for (linha = 0; linha < 3; linha++)
    if (atual.tabuleiro [linha] [0] == vez &&
	atual.tabuleiro [linha] [1] == vez &&
	atual.tabuleiro [linha] [2] == vez)
      return TRUE;

  /* Verificar as colunas */

  for (coluna = 0; coluna < 3; coluna++)
    if (atual.tabuleiro [0] [coluna] == vez &&
	atual.tabuleiro [1] [coluna] == vez &&
	atual.tabuleiro [2] [coluna] == vez)
      return TRUE;

  /* Verificar as diagonais */

  if (atual.tabuleiro [0] [0] == vez &&
      atual.tabuleiro [1] [1] == vez &&
      atual.tabuleiro [2] [2] == vez)
	return TRUE;

  if (atual.tabuleiro [0] [2] == vez &&
      atual.tabuleiro [1] [1] == vez &&
      atual.tabuleiro [2] [0] == vez)
	return TRUE;

  return FALSE;
}

/* gere_movimentos: Produz todas as combinacoes possiveis de jogada em uma
   dada profundidade.
   Recebe: A profundidade da arvore de pesquisa (de acordo com o algoritmo
   original do livro) para saber de quem e a vez de jogar, um tabuleiro
   indicando a situacao atual do jogo, uma estrutura por referencia para
   conter todas as novas possibilidades.
   Retorna: O numero de novas configuracoes geradas. */

int gere_movimentos (int profundidade, tipo_tabuleiro tabl_atual,
		     tipo_movimentos *movs)
{
  int linha,
      coluna,
      num_mov = 0;
  peca vez;

  /* Verificar de quem e a vez de jogar */

  if (PAR (profundidade))
    vez = CIRCULO;
  else
    vez = CRUZ;

  for (linha = 0; linha < 3; linha++)
    for (coluna = 0; coluna < 3; coluna++)
      if (tabl_atual.tabuleiro [linha] [coluna] == VAZIO) {
	int linha_aux,
	    coluna_aux;

	/* Copia tabuleiro atual para o vetor de movimentos plausiveis */

	for (linha_aux = 0; linha_aux < 3; linha_aux++)
	  for (coluna_aux = 0; coluna_aux < 3; coluna_aux++)
	    (*movs).movimentos [num_mov].tabuleiro [linha_aux] [coluna_aux] =
	    tabl_atual.tabuleiro [linha_aux] [coluna_aux];

	/* Marca o movimento possivel */

	(*movs).movimentos [num_mov].tabuleiro [linha] [coluna] = vez;

	num_mov++;
      }

  return num_mov;
}

/* estatico: Atribui uma nota de -9 a 9 a um tabuleiro de acordo com a
   profundidade da arvore de pesquisa.
   Recebe: O tabuleiro e a produndidade que ele se encontra.
   Retorna: Um inteiro correspondente a nota. */

int estatico (tipo_tabuleiro tabl, int profundidade)
{
  int linha,
      coluna;
  int total_cruz = 0,    /* Numero de possibilidades de vitoria da cruz */
      total_circulo = 0; /* Numero de possibilidades de vitoria do circulo */

  if (PAR (profundidade)) {
    if (venceu (CIRCULO, tabl))
      return (9);
    if (venceu (CRUZ, tabl))
      return (-9);
  }
  else {
    if (venceu (CRUZ, tabl))
      return (9);
    if (venceu (CIRCULO, tabl))
      return (-9);
  }

  /* Verificar as linhas */

  for (linha = 0; linha < 3; linha++) {
    if (tabl.tabuleiro [linha] [0] != CIRCULO &&
	tabl.tabuleiro [linha] [1] != CIRCULO &&
	tabl.tabuleiro [linha] [2] != CIRCULO)
      total_cruz++;
    if (tabl.tabuleiro [linha] [0] != CRUZ &&
	tabl.tabuleiro [linha] [1] != CRUZ &&
	tabl.tabuleiro [linha] [2] != CRUZ)
      total_circulo++;
  }

  /* Verificar as colunas */

  for (coluna = 0; coluna < 3; coluna++) {
    if (tabl.tabuleiro [0] [coluna] != CIRCULO &&
	tabl.tabuleiro [1] [coluna] != CIRCULO &&
	tabl.tabuleiro [2] [coluna] != CIRCULO)
      total_cruz++;
    if (tabl.tabuleiro [0] [coluna] != CRUZ &&
	tabl.tabuleiro [1] [coluna] != CRUZ &&
	tabl.tabuleiro [2] [coluna] != CRUZ)
      total_circulo++;
  }

  /* Verificar as diagonais */

  if (tabl.tabuleiro [0] [0] != CIRCULO &&
      tabl.tabuleiro [1] [1] != CIRCULO &&
      tabl.tabuleiro [2] [2] != CIRCULO)
    total_cruz++;
  if (tabl.tabuleiro [0] [0] != CRUZ &&
      tabl.tabuleiro [1] [1] != CRUZ &&
      tabl.tabuleiro [2] [2] != CRUZ)
    total_circulo++;

  if (tabl.tabuleiro [2] [0] != CIRCULO &&
      tabl.tabuleiro [1] [1] != CIRCULO &&
      tabl.tabuleiro [0] [2] != CIRCULO)
    total_cruz++;
  if (tabl.tabuleiro [2] [0] != CRUZ &&
      tabl.tabuleiro [1] [1] != CRUZ &&
      tabl.tabuleiro [0] [2] != CRUZ)
    total_circulo++;

  if (PAR (profundidade))
    return total_circulo - total_cruz;
  else
    return total_cruz - total_circulo;

}

/* minimax_alfa_beta: Implementacao do algoritmo recursivo de pesquisa em
   profundidade limitada utilizando cortes alfa e beta.
   Recebe: O tabuleiro a ser pesquisada a jogada, a profundidade inicial e
   dois inteiros, trilha_uso e trilha_pass, para a implementacao dos cortes
   alfa e beta.
   Retorna: Um tabuleiro com a jogada ja executada e a nota atribuida a esta
   jogada. */

int minimax_alfa_beta (tipo_tabuleiro posicao, int profundidade,
	int trilha_uso, int trilha_pass, tipo_tabuleiro *jogada)
{
  tipo_movimentos mov_plausiveis;  /* Configuracoes possiveis da
				      jogada atual */
  tipo_tabuleiro melhor_movimento; /* Melhor movimento ate agora */
  byte num_movs,                   /* Numero de tabuleiros possiveis
				      gerados a partir da situacao
				      atual */
       cont;


  if (PROFUNDIDADE_SUFICIENTE (profundidade)) {
    inicie (jogada);               /* Nao retorna nenhum tabuleiro */
    return estatico (posicao, profundidade);
  }

  num_movs = gere_movimentos (profundidade, posicao, &mov_plausiveis);

  if (num_movs == 0) {
    inicie (jogada);               /* Nao retorna nenhum tabuleiro */
    return estatico (posicao, profundidade);
  }

  for (cont = 0; cont < num_movs; cont++) {

    int result_suc,
	novo_valor;

    result_suc = minimax_alfa_beta (mov_plausiveis.movimentos [cont],
      profundidade + 1, -trilha_pass, -trilha_uso, &melhor_movimento);

    novo_valor = -result_suc;

    if (novo_valor > trilha_pass) {
      copie (mov_plausiveis.movimentos [cont], jogada);
      trilha_pass = novo_valor;
    }

    if (trilha_pass >= trilha_uso) {
      return (trilha_pass);
    }
  }
  return (trilha_pass);
}

main ()
{

  tipo_tabuleiro tabuleiro_atual, /* Tabuleiro principal das jogadas */
		 movimento;       /* Tabuleiro ja com o movimento do
					     algoritmo MINIMAX-ALFA-BETA */
  int linha,
      coluna;

  inicialize_video ();
  inicie (&tabuleiro_atual);

  do {
    receba_movimento (&tabuleiro_atual);
    if (venceu (CRUZ, tabuleiro_atual) || cheio (tabuleiro_atual))
      break;

    minimax_alfa_beta (tabuleiro_atual, 0, 10, -10, &movimento);
    coordenadas_movimento (&linha, &coluna, tabuleiro_atual, movimento);
    assinale_circulo (linha, coluna);
    copie (movimento, &tabuleiro_atual);
    if (venceu (CIRCULO, tabuleiro_atual) || cheio (tabuleiro_atual))
      break;

  } while (TRUE);

  mostre_tabuleiro (tabuleiro_atual);
}